ReadIni()
{

    local _inFile="${1:--}"
    local _section="$2"
    local _key="$3"
    local _outVar="${4:-}"

    local _cmd=''
    _cmd=$(cat <<'AWK_PROGRAM'
BEGIN {
    FS = "=";

    sectionFound = (section == "");
    keyFound = 0;
}

END {
    if (keyFound)
    {
        exit 0;
    }

    exit (sectionFound ? 102 : 101);
}

$0 == "[" section "]" {
    sectionFound = 1;
    next;
}

/^\[/ {
    if (sectionFound)
    {
        exit;
    }
}

sectionFound && $1 == key {
    keyFound = 1;
    print $2;
    exit;
}
AWK_PROGRAM
)
    local _value=''
    _value="$(awk -v section="${_section}" -v key="${_key}" -- "${_cmd}" "${_inFile}" 2>/dev/null)"
    local res=$?
    if [ ${res} -ne 0 ]; then
        return ${res}
    fi

    if [ -n "${_outVar}" ]; then
        eval "${_outVar#@}=\"\${_value}\""
    fi

    return 0
}

WriteIni()
{
    local _inFile="${1:--}"
    local _section="$2"
    local _key="$3"
    local _value="$4"
    local _outFile="${5:-${_inFile}}"

    local _cmd=''
    _cmd=$(cat <<'AWK_PROGRAM'
BEGIN {
    FS = "=";
    sectionFound = 0;
    keySet = 0;
    emptyLines = 0;
}

END {
    if (!keySet)
    {
        if (sectionFound)
        {
            print key "=" value;
            printEmptyLines();
        }
        else
        {
            print "[" section "]";
            print key "=" value;
            printEmptyLines();
        }
    }
}

function printTail(    s)
{
    while ((getline s) > 0)
    {
        print s;
    }
}

function printEmptyLines()
{
    while (emptyLines > 0)
    {
        print "";
        --emptyLines;
    }
}

$0 == "[" section "]" {
    sectionFound = 1;
    print;
    next;
}

/^\[/ {
    if (sectionFound)
    {
        print key "=" value;
        keySet = 1;
        printEmptyLines();
        print;
        printTail();
        exit;
    }
    else
    {
        print;
        next;
    }
}

sectionFound {
    if ($1 == key)
    {
        printEmptyLines();
        print $1 "=" value;
        keySet = 1;
        printTail();
        exit;
    }
    else
    {
        if ($0 == "")
        {
            ++emptyLines;
        }
        else
        {
            printEmptyLines();
            print;
        }
        next;
    }
}

{ print; }
AWK_PROGRAM
)
    local _tmpFilename=''
    local _redirTxt=''

    if [ "${_inFile}" = '-' ] || [ -f "${_inFile}" ]; then
        _redirTxt=' "${_inFile}" '
    else
        _redirTxt=' </dev/null '
    fi

    if [ "${_outFile}" != '-' ]; then
        _tmpFilename="${_outFile}.tmp";
        _redirTxt="${_redirTxt}"' 1>"${_tmpFilename}"'
    fi

    eval 'awk -v section="${_section}" -v key="${_key}" -v value="${_value}" -- "${_cmd}"' ${_redirTxt} 2>/dev/null
    local res=$?
    if [ ${res} -ne 0 ]; then
        [ -n "${_tmpFilename}" ] && rm -f "${_tmpFilename}" 2>/dev/null
    else
        [ -n "${_tmpFilename}" ] && { mv -T "${_tmpFilename}" "${_outFile}" 2>/dev/null; res=$?; }
    fi

    return ${res}
}
