GetUpgradeState()
{
    local stateVarName="${1:-}"
    local newVersionVarName="${2:-}"
    local oldVersionVarName="${3:-}"

    local current='/var/opt/kaspersky/kesl/install-current'
    local upgrade='/var/opt/kaspersky/kesl/install-upgrade'
    local previous='/var/opt/kaspersky/kesl/install-previous'
    local failed='/var/opt/kaspersky/kesl/install-failed'

    local manifestFilePath='/opt/kaspersky/kesl/doc/manifest.ini'
    local oldIniPath=''
    local newIniPath=''

    local state=''

    if [ ! -h "${current}" ]; then
        state='NotInstalled'
    else
        if [ ! -h "${failed}" ]; then
            if [ ! -h "${upgrade}" ]; then
                if [ ! -h "${previous}" ]; then
                    state='Installed'
                    oldIniPath="${current}"
                    newIniPath="${current}"
                else
                    state='Start'
                    oldIniPath="${previous}"
                    newIniPath="${current}"
                fi
            else
                if [ ! -h "${previous}" ]; then
                    state='Uninstall'
                    oldIniPath="${current}"
                    newIniPath="${upgrade}"
                else
                    state='Upgrade'
                    oldIniPath="${current}"
                    newIniPath="${upgrade}"
                fi
            fi
        else
            if [ ! -h "${previous}" ]; then
                if [ ! -h "${upgrade}" ]; then
                    state='RollingBack'
                    oldIniPath="${current}"
                    newIniPath="${failed}"
                else
                    state='RolledBack'
                    oldIniPath="${current}"
                    newIniPath="${failed}"
                fi
            else
                local curr
                if ! curr=$(readlink "${current}"); then
                    echo "Faild to read 'current' symlink" 1>&2
                    return 255
                fi

                local prev
                if ! prev=$(readlink "${previous}"); then
                    echo "Faild to read 'previous' symlink" 1>&2
                    return 255
                fi

                if [ "${curr}" = "${prev}" ]; then
                    state='UpgradeFailed'
                    oldIniPath="${current}"
                    newIniPath="${failed}"
                else
                    state='StartFailed'
                    oldIniPath="${previous}"
                    newIniPath="${failed}"
                fi
            fi
        fi

        local grepRegEx='^\s*ProductVersion\s*='
        local sedCmd='s:\(.*=\s*\)\(\S*\)\(\s*\):\2:'
        local ver=''

        if [ ! -z "${newVersionVarName}" ]; then
            ver=''
            if [ ! -z "${newIniPath}" ]; then
                if ver=$(grep "${grepRegEx}" "${newIniPath}/${manifestFilePath}"); then
                    ver=$(echo "${ver}" | sed -e "${sedCmd}")
                fi
            fi
            eval "${newVersionVarName#@}=\"\${ver}\""
        fi

        if [ ! -z "${oldVersionVarName}" ]; then
            ver=''
            if [ ! -z "${oldIniPath}" ]; then
                if ver=$(grep "${grepRegEx}" "${oldIniPath}/${manifestFilePath}"); then
                    ver=$(echo "${ver}" | sed -e "${sedCmd}")
                fi
            fi
            eval "${oldVersionVarName#@}=\"\${ver}\""
        fi
    fi

    if [ -z "${stateVarName}" ]; then
        echo "${state}"
    else
        eval "${stateVarName#@}=\"\${state}\""
    fi

    return 0
}
