package KL::PlaformSpecificUtils;

use strict;

use base 'KL::AppBase';

use POSIX qw(setlocale);
BEGIN
{
    my $lang = substr(setlocale("LC_ALL"), 0, 2);
    my @LFS_LANGS = split(/\s*,\s*/, "%LFS_LICENSE_LANGS%");
    if (grep {$_ eq $lang} @LFS_LANGS)
    {
        eval "require KL::Localization::$lang; KL::Localization::$lang->import();";
    }
    else
    {
        require KL::Localization::en;
        KL::Localization::en->import();
    }
}

sub new
{
    my $class = shift;
    my $appid = shift;

    my $self = $class->SUPER::new($appid);

    push @{$self->{'COMMANDS'}}, {
        'id' => 'build_oas',
        'spec' => 'b|build:s',
        'sub' => 'do_build_oas_module',
        'default' => undef,
        'descr' => $MSG_CMD_HELP_BUILD,
    };
    
    return $self;
}

sub do_build_oas_module
{
    my $self = shift;
    my $args = shift;

    KL::Utils::output($HEADER_SETUP_OAS);

    my $setup_cmd = $self->{'SETUP'};
    my $srcs_dir = $self->get_path('srcs') . '/kernel';

    if (! -f "$srcs_dir/configure")
    {
        KL::Utils::warning($MSG_OAS_CONFIGURE_NOT_FOUND.$MSG_OAS_MODULE_WARN,
                           $setup_cmd, $srcs_dir);
        return 0;
    }

    my @build_tools = ('make', 'gcc');
    my $install_result = 0;

    for (my $try_counter = 0; $try_counter < 2; $try_counter++)
    {
        my @absent_build_tools;

        foreach my $t (@build_tools)
        {
            if (!KL::Utils::which($t))
            {
                push @absent_build_tools, $t;
            }
        }

        last if (not @absent_build_tools);

        my $atools = join("\n    ", "", @absent_build_tools);

        if ($try_counter > 0)
        {
            if ($install_result != 0)
            {
                KL::Utils::warning($MSG_OAS_BUILD_TOOLS_INSTALLATION_FAILED) ;
            }
            else
            {
                KL::Utils::warning($MSG_OAS_BUILD_TOOLS_NOT_FOUND);
                print "$atools\n";
            }

            return 0;
        }

        KL::Utils::warning($MSG_OAS_BUILD_TOOLS_NOT_FOUND);
        print "$atools\n";
        KL::Utils::output($MSG_OAS_NEED_INSTALL_NECESSARY_PACKAGES);

        my $cmd = $self->get_build_tools_install_cmd();
        if (!$cmd)
        {
            KL::Utils::output($MSG_OAS_BUILD_TOOLS_DONT_KNOWN_HOW_TODO_AUTO_INSTALL);
            return 0;
        }

        if ($self->ai_mode)
        {
            if (lc($self->ai_get_value('KERNEL_SRCS_INSTALL')) eq 'no')
            {
                return 0;
            }
        }
        elsif (!KL::Utils::ask_boolean('y', $MSG_OAS_AUTO_INSTALL_REQUIRED_PACKAGES_Q))
        {
            return 0;
        }


        $install_result = system($cmd);
    }

    my $kernel_srcs = $args->{'build_oas'};
    if (!$kernel_srcs)
    {
        $install_result = 0;
        for (my $try_counter = 0; $try_counter < 2; $try_counter++)
        {
            $kernel_srcs = $self->find_kernel_sources();

            last if ($kernel_srcs);

            if ($try_counter > 0)
            {
                if ($install_result != 0)
                {
                    KL::Utils::warning($MSG_OAS_LINUX_SRC_INSTALLATION_FAILED);
                }
                else
                {
                    KL::Utils::warning($MSG_OAS_KERNEL_SRCS_NOT_FOUND);
                }

                return 0;
            }

            KL::Utils::warning($MSG_OAS_KERNEL_SRCS_NOT_FOUND);
            KL::Utils::output($MSG_OAS_KERNEL_SRCS_PROBABLY_NEED_INSTALL);

            my $cmd = $self->get_kernel_headers_install_cmd();
            if (!$cmd)
            {
                KL::Utils::warning($MSG_OAS_LINUX_SRC_DONT_KNOWN_HOW_TODO_AUTO_INSTALL);
                return 0;
            }

            if ($self->ai_mode)
            {
                if (lc($self->ai_get_value('KERNEL_SRCS_INSTALL')) eq 'no')
                {
                    return 0;
                }
            }
            elsif (!KL::Utils::ask_boolean('y', $MSG_OAS_AUTO_INSTALL_REQUIRED_PACKAGES_Q))
            {
                return 0;
            }

            $self->ai_save('KERNEL_SRCS_INSTALL', 'Yes');

            KL::Utils::output("<B>$cmd</B>\n");

            $install_result = system($cmd);
        }
    }

    KL::Utils::output($MSG_OAS_LINUX_SRC_FOUND_AT, $kernel_srcs);

    if (!KL::Utils::do_in_fork( sub { $self->make_linux_module($kernel_srcs, $args->{'kernel_module'}); } ))
    {
        KL::Utils::warning($MSG_OAS_MODULE_WARN, $setup_cmd);
        return 0;
    }

    return 1;
}

1;
